#ifdef __cplusplus
extern "C" {
#endif
#include <ctype.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef __cplusplus
}
#endif

#include <sysexits.h>
#include <CoreFoundation/CFSocket.h>
#include "ServiceUSBToolbox.h"

enum
{
	e_set_output=1,
	e_get_output_value,
	e_get_input_value,
	e_get_ex_value,
	e_get_ey_value,
	e_vendor_request,
	e_byte_count,
	e_read_bytes,
	e_write_bytes,
	e_release_device,
	
	e_set_output_by_mask,
	e_set_output_bit,
	e_get_input_value_as_mask,
	e_get_input_value_bit,
	e_test_input_value,
	e_compare_input_value,
	
	e_device_count,
	e_device_identifier,
	e_device_serialnr,
	e_device_ROM_version,
	e_device_driver_version,
	
	e_help
};

typedef struct {
	char	*cmd;
	int	minArgs;
	int	maxArgs;
	int	commandID;
	char	*usage;
} cmdInfo;

const cmdInfo commands[] = {
	// cmd		minArgs	maxArgs	commandID
	// 	usage	

	{ "help",0,0,e_help,
		" help                                          : list available commands"			},

	{ "set_output",1,2,e_set_output,
		" set_output value [deviceID]                   : sets the output value"			},

	{ "get_output_value",0,1,e_get_output_value,
		" get_output_value [deviceID]                   : return the output value"			},

	{ "get_input_value",0,1,e_get_input_value,
		" get_input_value [deviceID]                    : return the input value"			},

	{ "get_ex_value",0,1,e_get_ex_value,
		" get_ex_value [deviceID]                       : return the ex value"			},

	{ "get_ey_value",0,1,e_get_ey_value,
		" get_ey_value [deviceID]                       : returns the ey value"			},

	{ "vendor_request",3,4,e_vendor_request,
		" vendor_request request value index [deviceID] : sends a vendor request"			},

	{ "byte_count",1,2,e_byte_count,
		" byte_count pipe [deviceID]                    : return the number of bytes in the specified pipe"			},

	{ "read_bytes",1,3,e_read_bytes,
		" read_bytes pipe [count] [deviceID]            : reads the number of bytes from the specified pipe. If count is not specified."			},

	{ "write_bytes",2,3,e_write_bytes,
		" write_bytes pipe bytes [deviceID]             : write the specified bytes to the pipe"			},

	{ "release_device",0,1,e_release_device,
		" release_device [deviceID]                     : releases the device so that it available to the apps"			},



	{ "set_output_by_mask",1,2,e_set_output_by_mask,
		" set_output_by_mask mask [deviceID]            : set the output value using a mask"			},

	{ "set_output_bit",2,3,e_set_output_bit,
		" set_output_bit symbol bit [deviceID]          : set the output value useing a bit"			},

	{ "get_input_value_as_mask",0,1,e_get_input_value_as_mask,
		" get_input_value_as_mask [deviceID]            : get input value as a mask"			},

	{ "get_input_value_bit",1,2,e_get_input_value_bit,
		" get_input_value_bit bit [deviceID]            : get input value using a bit"			},

	{ "test_input_value",1,2,e_test_input_value,
		" test_input_value bit [deviceID]               : test input value"			},

	{ "compare_input_value",1,2,e_compare_input_value,
		" compare_input_value mask [deviceID]           : compare input value against a mask"			},



	{ "device_count",0,1,e_device_count,
		" device_count                                  : returns the number of devices available"			},

	{ "device_identifier",0,1,e_device_identifier,
		" device_identifier [last_deviceID]             : returns the identifier of the next device. If last_deviceID is not specified the first one will be returned."			},

	{ "device_serialnr",0,1,e_device_serialnr,
		" device_serialnr [deviceID]                    : returns the device serialnr."			},

	{ "device_ROM_version",0,1,e_device_ROM_version,
		" device_ROM_version [deviceID]                 : returns the ROM version"			},

	{ "device_driver_version",0,1,e_device_driver_version,
		" device_driver_version [deviceID]              : returns the device driver version"			},

};

const int nCommands = (sizeof(commands)/sizeof(cmdInfo));

#define LINE_LENGTH 256

static int			nesting		= 0;
static CFMutableArrayRef	sources		= NULL;

static UInt16 GetCommandSelector(char* inCommandString);
static void usage(void);
static char * GetErrorText(OSErr Ooops);
static void runLoopProcessInput(CFSocketRef s, CFSocketCallBackType type, CFDataRef address, const void *data, void *info);
static void execute_command(int argc,char * argv[],Boolean CommandLineState);
static char *getString(char **line);
static char *getLine(char *buf, int len, FILE *fp);
static void HandleError(OSErr Ooops,Boolean CommandLineState);

int main(int argc,char * argv[])
{
	if (argc==1)
	{
		CFSocketContext			context	= { 0, stdin, NULL, NULL, NULL };
		CFRunLoopSourceRef	rls;
		
		// create a "socket" reference with the file descriptor associated with stdin
		CFSocketRef in=CFSocketCreateWithNative(NULL,STDIN_FILENO,kCFSocketReadCallBack, runLoopProcessInput,&context);

		// Create a run loop source for the (stdin) file descriptor
		rls = CFSocketCreateRunLoopSource(NULL,in,nesting);
		
		// keep track of input file sources
		sources = CFArrayCreateMutable(NULL,0,&kCFTypeArrayCallBacks);
		CFArrayAppendValue(sources, rls);

		// add this source to the run loop
		CFRunLoopAddSource(CFRunLoopGetCurrent(),rls,kCFRunLoopDefaultMode);
		CFRelease(rls);
		CFRelease(in);

		// issue (initial) prompt
		if (isatty(STDIN_FILENO))
		{
			printf("> ");
			fflush(stdout);
		}

		CFRunLoopRun();	// process input, process events

		exit (EX_OK);	// insure the process exit status is 0
		return 0;	// ...and make main fit the ANSI spec.
	}
	
	execute_command(argc,argv,false);
	
	exit(0);
}

static void execute_command(int argc,char * argv[],Boolean CommandLineState)
{
	char									command[256];
	UInt16								CommandID=0;
	OSErr									Ooops=noErr;
	ServiceUSB_DeviceIdentifierRef		DeviceIdentifier=0;
	Boolean								ValueValid=false;
	Boolean								IndexValid=false;
	Boolean								PipeIndexValid=false;
	Boolean								VendorRequestValid=false;
	UInt32								Value=0;
	UInt32								Index=0;
	UInt32								VendorRequest=0;
	UInt8									PipeIndex=0xFF;
	UInt32								ReadByteCount=0;
	char*									ByteStream=NULL;
	UInt32								ByteStreamCount=0;
	ServiceUSB_BitMaskStringType	theMask;
	char									theSymbol;
	UInt8									theBit;

	memset(command,0,sizeof(command));
	
	if (CommandLineState)
	{
		if (argc>0)
		{
			strcpy(command,argv[0]);
			
			int	i;
			char	*cmd = argv[0];

			for (i=0; i<nCommands; i++)
			{
				if (strcasecmp(cmd, commands[i].cmd) == 0)
				{
					--argc;
					argv++;
					if (argc < commands[i].minArgs)
					{
						printf("%s: too few arguments\n",cmd);
						return;
					}
					else if (argc > commands[i].maxArgs)
					{
						printf("%s: too many arguments\n",cmd);
						return;
					}
					
					CommandID=commands[i].commandID;
				}
			}
		}
		else return;

		switch(CommandID)
		{
			case e_set_output:
				Value=atoi(argv[0]);
				ValueValid=true;
				--argc; argv++;
				
				if (argc>0) DeviceIdentifier=(ServiceUSB_DeviceIdentifierRef)atoi(argv[0]);
			break;
			case e_get_output_value:
				if (argc>0) DeviceIdentifier=(ServiceUSB_DeviceIdentifierRef)atoi(argv[0]);
			break;
			case e_get_input_value:
				if (argc>0) DeviceIdentifier=(ServiceUSB_DeviceIdentifierRef)atoi(argv[0]);
			break;
			case e_get_ex_value:
				if (argc>0) DeviceIdentifier=(ServiceUSB_DeviceIdentifierRef)atoi(argv[0]);
			break;
			case e_get_ey_value:
				if (argc>0) DeviceIdentifier=(ServiceUSB_DeviceIdentifierRef)atoi(argv[0]);
			break;
			case e_vendor_request:
				VendorRequest=atoi(argv[0]);
				VendorRequestValid=true;
				--argc; argv++;

				Value=atoi(argv[0]);
				ValueValid=true;
				--argc; argv++;

				Index=atoi(argv[0]);
				IndexValid=true;
				--argc; argv++;

				if (argc>0) DeviceIdentifier=(ServiceUSB_DeviceIdentifierRef)atoi(argv[0]);
			break;
			case e_byte_count:
				PipeIndex=atoi(argv[0]);
				PipeIndexValid=true;
				--argc; argv++;
				
				if (argc>0) DeviceIdentifier=(ServiceUSB_DeviceIdentifierRef)atoi(argv[0]);
			break;
			case e_read_bytes:
				PipeIndex=atoi(argv[0]);
				PipeIndexValid=true;
				--argc; argv++;

				if (argc>0)
				{
					ReadByteCount=atoi(argv[0]);
					--argc; argv++;
				}
					
				if (argc>0) DeviceIdentifier=(ServiceUSB_DeviceIdentifierRef)atoi(argv[0]);
			break;
			case e_write_bytes:
				PipeIndex=atoi(argv[0]);
				PipeIndexValid=true;
				--argc; argv++;
				
				ByteStreamCount=strlen(argv[0]);
				if (ByteStreamCount>0)
				{
					ByteStream=(char*)malloc(ByteStreamCount);
					if (ByteStream==NULL) exit(1); // should never happen
					strcpy(ByteStream,argv[0]);
				}
				--argc; argv++;

				if (argc>0) DeviceIdentifier=(ServiceUSB_DeviceIdentifierRef)atoi(argv[0]);
			break;
			case e_release_device:
				if (argc>0) DeviceIdentifier=(ServiceUSB_DeviceIdentifierRef)atoi(argv[0]);
			break;


			case e_set_output_by_mask:
				if (strlen(argv[0])!=8)
				{
					fprintf(stderr,"Error: Mask need to be 8 characters.\n");
					return;
				}
				memcpy(theMask,argv[0],strlen(argv[0]));
				--argc; argv++;

				if (argc>0) DeviceIdentifier=(ServiceUSB_DeviceIdentifierRef)atoi(argv[0]);
			break;
			case e_set_output_bit:
				if (strlen(argv[0])!=1)
				{
					fprintf(stderr,"Error: Symbol need to be 1 character.\n");
					return;
				}
				memcpy(&theSymbol,argv[0],strlen(argv[0]));
				--argc; argv++;

				theBit=atoi(argv[0]);
				--argc; argv++;

				if (argc>0) DeviceIdentifier=(ServiceUSB_DeviceIdentifierRef)atoi(argv[0]);
			break;
			case e_get_input_value_as_mask:
				if (argc>0) DeviceIdentifier=(ServiceUSB_DeviceIdentifierRef)atoi(argv[0]);
			break;
			case e_get_input_value_bit:
				theBit=atoi(argv[0]);
				--argc; argv++;

				if (argc>0) DeviceIdentifier=(ServiceUSB_DeviceIdentifierRef)atoi(argv[0]);
			break;
			case e_test_input_value:
				theBit=atoi(argv[0]);
				--argc; argv++;

				if (argc>0) DeviceIdentifier=(ServiceUSB_DeviceIdentifierRef)atoi(argv[0]);
			break;
			case e_compare_input_value:
				if (strlen(argv[0])!=8)
				{
					fprintf(stderr,"Error: Mask need to be 8 characters.\n");
					return;
				}
				memcpy(theMask,argv[0],strlen(argv[0]));
				--argc; argv++;

				if (argc>0) DeviceIdentifier=(ServiceUSB_DeviceIdentifierRef)atoi(argv[0]);
			break;


			case e_device_count:
			break;
			case e_device_identifier:
				if (argc>0) DeviceIdentifier=(ServiceUSB_DeviceIdentifierRef)atoi(argv[0]);
			break;
			case e_device_serialnr:
				if (argc>0) DeviceIdentifier=(ServiceUSB_DeviceIdentifierRef)atoi(argv[0]);
			break;
			case e_device_ROM_version:
				if (argc>0) DeviceIdentifier=(ServiceUSB_DeviceIdentifierRef)atoi(argv[0]);
			break;
			case e_device_driver_version:
				if (argc>0) DeviceIdentifier=(ServiceUSB_DeviceIdentifierRef)atoi(argv[0]);
			break;
		}
	}
	else
	{
		char 				ch;
		extern char *optarg;
		extern int  optind;

		while ((ch = getopt(argc, argv, "c:v:d:i:p:b:r:n:m:s:t:")) != -1)
		{
			switch (ch)
			{
				case 'c':
					strcpy((char*)command,optarg);
				break;
				case 'v':
					Value=atoi(optarg);
					ValueValid=true;
				break;
				case 'd':
					DeviceIdentifier=(ServiceUSB_DeviceIdentifierRef)atoi(optarg);
				break;
				case 'i':
					Index=atoi(optarg);
					IndexValid=true;
				break;
				case 'r':
					VendorRequest=atoi(optarg);
					VendorRequestValid=true;
				break;
				case 'p':
					PipeIndex=atoi(optarg);
					PipeIndexValid=true;
				break;
				case 'n':
					ReadByteCount=atoi(optarg);
				break;
				case 'b':
					ByteStreamCount=strlen(optarg);
					if (ByteStreamCount>0)
					{
						ByteStream=(char*)malloc(ByteStreamCount);
						if (ByteStream==NULL) exit(1); // should never happen
						strcpy(ByteStream,optarg);
					}
				break;
				case 'm':
					if (strlen(optarg)!=8)
					{
						fprintf(stderr,"Error: Mask need to be 8 characters.\n");
						if (!CommandLineState) exit(1);
						else return;
					}
					memcpy(theMask,optarg,strlen(optarg));
				break;
				case 's':
					if (strlen(optarg)!=1)
					{
						fprintf(stderr,"Error: Symbol need to be 1 character.\n");
						if (!CommandLineState) exit(1);
						else return;
					}
					memcpy(&theSymbol,optarg,strlen(optarg));
				break;
				case 't':
					theBit=atoi(optarg);
				break;
				
				default: usage();
			}
		}
		argc -= optind;
		argv += optind;
		
		CommandID=GetCommandSelector(command);
	}
	

	switch(CommandID)
	{
		case e_set_output:
			if (!ValueValid) usage();
			if (Value<0 || Value>255)
			{
				fprintf(stderr,"Error: Value needs to be from 0 to 255.\n");
				if (!CommandLineState) exit(1);
			}
			else
			{
				Ooops=ServiceUSB_SetOutputValue(Value,DeviceIdentifier);
				HandleError(Ooops,CommandLineState);
			}
		break;
		case e_get_output_value:
			printf("%d\n",ServiceUSB_GetOutputValue(DeviceIdentifier));
		break;
		case e_get_input_value:
			printf("%d\n",ServiceUSB_GetInputValue(DeviceIdentifier));
		break;
		case e_get_ex_value:
			printf("%d\n",ServiceUSB_GetEXValue(DeviceIdentifier));
		break;
		case e_get_ey_value:
			printf("%d\n",ServiceUSB_GetEYValue(DeviceIdentifier));
		break;
		case e_vendor_request:
			if (!ValueValid || !IndexValid || !VendorRequestValid) usage();
			Ooops=ServiceUSB_SendVendorRequest(VendorRequest,Value,Index,DeviceIdentifier);
			HandleError(Ooops,CommandLineState);
		break;
		case e_byte_count:
		{
			if (!PipeIndexValid) usage();
			if (Value>2)
			{
				fprintf(stderr,"Error: Pipe needs to be from 0 to 2.\n");
				if (!CommandLineState) exit(1);
			}
			printf("%d\n",(int)ServiceUSB_GetBytesInReadBuffer(PipeIndex,DeviceIdentifier));
		}
		break;
		case e_read_bytes:
		{
			if (Value>2)
			{
				fprintf(stderr,"Error: Pipe needs to be from 0 to 2.\n");
				if (!CommandLineState) exit(1);
			}
			if (ReadByteCount==0) ReadByteCount=ServiceUSB_GetBytesInReadBuffer(PipeIndex,DeviceIdentifier);
			char* theBuffer=(char*)malloc(ReadByteCount+1);
			memset(theBuffer,0,ReadByteCount+1);
			ServiceUSB_ReadBytesFromReadBuffer(PipeIndex,theBuffer,&ReadByteCount,DeviceIdentifier);
			theBuffer[ReadByteCount]=0;
			printf("%s\n",theBuffer);
		}
		break;
		case e_write_bytes:
			if (Value>2)
			{
				fprintf(stderr,"Error: Pipe needs to be from 0 to 2.\n");
				if (!CommandLineState) exit(1);
			}

			if (ByteStreamCount>0 && ByteStream) Ooops=ServiceUSB_WriteBytes(PipeIndex,ByteStream,&ByteStreamCount,DeviceIdentifier);
			
			if (ByteStream) free(ByteStream);
			HandleError(Ooops,CommandLineState);
		break;
		

		case e_set_output_by_mask:
			Ooops=ServiceUSB_SetOutputValueByMask(theMask,DeviceIdentifier);
			HandleError(Ooops,CommandLineState);
		break;
		case e_set_output_bit:
			Ooops=ServiceUSB_SetOutputValueBit(theSymbol,theBit,DeviceIdentifier);
			HandleError(Ooops,CommandLineState);
		break;
		case e_get_input_value_as_mask:
			ServiceUSB_GetInputValueAsMask(theMask,DeviceIdentifier);
			printf("%.8s\n",theMask);
		break;
		case e_get_input_value_bit:
			printf("%c\n",ServiceUSB_GetInputValueBit(theBit,DeviceIdentifier));
		break;
		case e_test_input_value:
			printf("%d\n",ServiceUSB_TestInputValueBit(theBit,DeviceIdentifier));
		break;
		case e_compare_input_value:
			printf("%d\n",ServiceUSB_CompareInputValueMask(theMask,DeviceIdentifier));
		break;


		case e_device_count:
			printf("%d\n",ServiceUSB_GetDeviceCount());
		break;
		case e_device_identifier:
			printf("%d\n",(int)ServiceUSB_GetNextDeviceIdentifier(DeviceIdentifier));
		break;
		case e_device_serialnr:
		{
			Str255 Serialnr;
			
			Ooops=ServiceUSB_GetDeviceSerialnumber(Serialnr,DeviceIdentifier);
			HandleError(Ooops,CommandLineState);
			if (Ooops==noErr)
			{
				Serialnr[Serialnr[0]+1]=0;
				printf("%s\n",&Serialnr[1]);
			}
		}
		break;
		case e_device_ROM_version:
		{
			printf("0x%x\n",ServiceUSB_GetDeviceROMVersion(DeviceIdentifier));
		}
		break;
		case e_device_driver_version:
			Str255	VersionString;
			UInt32	VersionNum;
			
			VersionNum=ServiceUSB_GetUSBDriverVersion(DeviceIdentifier);
			ServiceUSB_Version2String(*(NumVersion*)&VersionNum,VersionString);
			VersionString[VersionString[0]+1]=0;
			printf("%s\n",&VersionString[1]);
		break;
		case e_help:
			printf("\nAvailable commands:\n");
			
			for (int i=0; i<nCommands; i++)
				printf("%s\n", commands[i].usage);
			
			printf("\n");
		break;
		default:
			printf("%s: unknown, type \"help\" for command info\n",command);
	}
}

static void HandleError(OSErr Ooops,Boolean CommandLineState)
{
	if (Ooops!=noErr)
	{
		fprintf(stderr,"Error: %s\n",GetErrorText(Ooops));
		if (!CommandLineState) exit(1);
	}
}

static void usage(void)
{
	fprintf(stderr, "usage: SERVICE_USB_Tool\n");
	fprintf(stderr, "or:    SERVICE_USB_Tool -c set_output -v value [-d deviceID]\n");
	fprintf(stderr, "       SERVICE_USB_Tool -c get_output_value [-d deviceID]\n");
	fprintf(stderr, "       SERVICE_USB_Tool -c get_input_value [-d deviceID]\n");
	fprintf(stderr, "       SERVICE_USB_Tool -c get_ex_value [-d deviceID]\n");
	fprintf(stderr, "       SERVICE_USB_Tool -c get_ey_value [-d deviceID]\n");
	fprintf(stderr, "       SERVICE_USB_Tool -c vendor_request -r request -v value -i index [-d deviceID]\n");
	fprintf(stderr, "       SERVICE_USB_Tool -c byte_count -p pipe [-d deviceID]\n");
	fprintf(stderr, "       SERVICE_USB_Tool -c read_bytes -p pipe [-n count] [-d deviceID]\n");
	fprintf(stderr, "       SERVICE_USB_Tool -c write_bytes -p pipe -b bytes [-d deviceID]\n");
	fprintf(stderr, "       SERVICE_USB_Tool -c release_device [-d deviceID]\n");

	fprintf(stderr, "       SERVICE_USB_Tool -c set_output_by_mask -m mask [-d deviceID]\n");
	fprintf(stderr, "       SERVICE_USB_Tool -c set_output_bit -s symbol -t bit [-d deviceID]\n");
	fprintf(stderr, "       SERVICE_USB_Tool -c get_input_value_as_mask [-d deviceID]\n");
	fprintf(stderr, "       SERVICE_USB_Tool -c get_input_value_bit -t bit [-d deviceID]\n");
	fprintf(stderr, "       SERVICE_USB_Tool -c test_input_value -t bit [-d deviceID]\n");
	fprintf(stderr, "       SERVICE_USB_Tool -c compare_input_value -m mask [-d deviceID]\n");
	
	fprintf(stderr, "       SERVICE_USB_Tool -c device_count\n");
	fprintf(stderr, "       SERVICE_USB_Tool -c device_identifier [-d last_deviceID]\n");
	fprintf(stderr, "       SERVICE_USB_Tool -c device_serialnr [-d deviceID]\n");
	fprintf(stderr, "       SERVICE_USB_Tool -c device_ROM_version [-d deviceID]\n");
	fprintf(stderr, "       SERVICE_USB_Tool -c device_driver_version [-d deviceID]\n");
	exit(1);
}

static UInt16 GetCommandSelector(char* inCommandString)
{
	if (strcmp(inCommandString,"set_output")==0) return e_set_output;
	if (strcmp(inCommandString,"get_output_value")==0) return e_get_output_value;
	if (strcmp(inCommandString,"get_input_value")==0) return e_get_input_value;
	if (strcmp(inCommandString,"get_ex_value")==0) return e_get_ex_value;
	if (strcmp(inCommandString,"get_ey_value")==0) return e_get_ey_value;
	if (strcmp(inCommandString,"vendor_request")==0) return e_vendor_request;
	if (strcmp(inCommandString,"byte_count")==0) return e_byte_count;
	if (strcmp(inCommandString,"read_bytes")==0) return e_read_bytes;
	if (strcmp(inCommandString,"write_bytes")==0) return e_write_bytes;
	if (strcmp(inCommandString,"release_device")==0) return e_release_device;
	
	if (strcmp(inCommandString,"set_output_by_mask")==0) return e_set_output_by_mask;
	if (strcmp(inCommandString,"set_output_bit")==0) return e_set_output_bit;
	if (strcmp(inCommandString,"get_input_value_as_mask")==0) return e_get_input_value_as_mask;
	if (strcmp(inCommandString,"get_input_value_bit")==0) return e_get_input_value_bit;
	if (strcmp(inCommandString,"test_input_value")==0) return e_test_input_value;
	if (strcmp(inCommandString,"compare_input_value")==0) return e_compare_input_value;
	
	if (strcmp(inCommandString,"device_count")==0) return e_device_count;
	if (strcmp(inCommandString,"device_identifier")==0) return e_device_identifier;
	if (strcmp(inCommandString,"device_serialnr")==0) return e_device_serialnr;
	if (strcmp(inCommandString,"device_ROM_version")==0) return e_device_ROM_version;
	if (strcmp(inCommandString,"device_driver_version")==0) return e_device_driver_version;
	return 0;
}

static Boolean process_line(FILE *fp)
{
	char	line[LINE_LENGTH], *s, *arg, **argv = NULL;
	int	i, argc;

	// if end-of-file, exit
	if (getLine(line, sizeof(line), fp) == NULL) return FALSE;

	if (nesting > 0) printf("%d> %s", nesting, line);

	// if requested, exit
	if (strcasecmp(line, "exit") == 0) return FALSE;
	if (strcasecmp(line, "quit") == 0) return FALSE;
	if (strcasecmp(line, "q"   ) == 0) return FALSE;

	// break up the input line
	s = line;
	argc = 0;
	while ((arg = getString(&s)) != NULL)
	{
		if (argc == 0)
			argv = (char **)malloc(2 * sizeof(char *));
		else
			argv = (char **)realloc(argv, ((argc + 2) * sizeof(char *)));
		argv[argc++] = arg;
	}

	// process the command
	if (argc > 0)
	{
		argv[argc] = NULL;	// just in case...
		char Command[256];

		memset(Command,0,sizeof(Command));
		if (*argv[0] != '#') execute_command(argc, argv,true);

		for (i = 0; i < argc; i++)
			free(argv[i]);
		free(argv);
	}

	return TRUE;
}

static void runLoopProcessInput(CFSocketRef s, CFSocketCallBackType type, CFDataRef address, const void *data, void *info)
{
	FILE* fp = (FILE*)info;

	if (process_line(fp) == FALSE)
	{
		// we don't want any more input from this stream, stop listening
		CFRunLoopRemoveSource(CFRunLoopGetCurrent(),(CFRunLoopSourceRef)CFArrayGetValueAtIndex(sources, 0),kCFRunLoopDefaultMode);

		// we no longer need the fd (socket)
		CFSocketInvalidate(s);

		// we no longer need to track this source
		CFArrayRemoveValueAtIndex(sources, 0);

		if (CFArrayGetCount(sources) > 0)
		{
			// add the previous input source to the run loo
			CFRunLoopAddSource(CFRunLoopGetCurrent(),(CFRunLoopSourceRef)CFArrayGetValueAtIndex(sources, 0), kCFRunLoopDefaultMode);
		}
		else exit (EX_OK); // no more input sources, we're done!

		
		nesting--; // decrement the nesting level
	}

	// if necessary, re-issue prompt
	if ((CFArrayGetCount(sources) == 1) && isatty(STDIN_FILENO))
	{
		printf("> ");
		fflush(stdout);
	}

	return;
}

static char * GetErrorText(OSErr Ooops)
{
	switch(Ooops)
	{
		case errServiceUSB_OutOfParamBlocks:
			return "No USB control blocks available. There are too many request executing.";
		break;
		case errServiceUSB_Timeout:
			return "A call did not finish for some reasons: Timeout was reached.";
		break;
		case errServiceUSB_BadSync:
			return "Call was issued from the interrupt level.";
		break;
		case errServiceUSB_USBDriverNotFound:
			return "Couldn't find the USB SERVICE driver. Probably it is not loaded.";
		break;
		case errServiceUSB_WrongPipeIndex:
			return "A wrong index for the read/write pipe was specified (valid values are 0, 1 and 2.";
		break;
		case errServiceUSB_DeviceNotFound:
			return "The device was not found. It might be in use by another application.";
		break;
		case errServiceUSB_CFNil:
			return "A call to core foundation returned NIL.";
		break;
		default: 
		{
			static char theString[256];
			
			sprintf(theString,"Unkown error %d (0x%x)",Ooops,Ooops);
			return theString;
		}
	}
}

static char *getLine(char *buf, int len, FILE *fp)
{
	int x;

	if (fgets(buf, len, fp) == NULL)
		return NULL;

	x = strlen(buf);
	if (buf[x-1] == '\n') {
		// the entire line fit in the buffer, remove the newline
		buf[x-1] = '\0';
	} else {
		// eat the remainder of the line
		do {
			x = fgetc(fp);
		} while ((x != '\n') && (x != EOF));
	}

	return buf;
}

static char *getString(char **line)
{
	char *s, *e, c, *string;
	int i, isQuoted = 0, escaped = 0;

	if (*line == NULL) return NULL;
	if (**line == '\0') return NULL;

	// Skip leading white space
	while (isspace(**line)) *line += 1;

	// Grab the next string
	s = *line;
	if (*s == '\0') {
		return NULL;				// no string available
	} else if (*s == '"') {
		isQuoted = 1;				// it's a quoted string
		s++;
	}

	for (e = s; (c = *e) != '\0'; e++) {
		if (isQuoted && (c == '"'))
			break;				// end of quoted string
		if (c == '\\') {
			e++;
			if (*e == '\0')
				break;			// if premature end-of-string
			if ((*e == '"') || isspace(*e))
				escaped++;		// if escaped quote or white space
		}
		if (!isQuoted && isspace(c))
			break;				// end of non-quoted string
	}

	string = (char*)malloc(e - s - escaped + 1);

	for (i = 0; s < e; s++) {
		string[i] = *s;
		if (!((s[0] == '\\') && ((s[1] == '"') || isspace(s[1])))) i++;
	}
	string[i] = '\0';

	if (isQuoted)
		e++;					// move past end of quoted string

	*line = e;
	return string;
}
